
package Extract;
;# ========================================================================
;# Extract.pm -- extract important words from text
;#
;# last modified : 020116
;# version       : 1.15
;#
;# by Keisuke Uchima 
;#    Hirokazu Ohata
;#    Hiroaki  Yumoto (Email:hir@forest.dnj.ynu.ac.jp)
;# ========================================================================

=head1 NAME

    Extract -- ׸Х⥸塼

=head1 SYNOPSIS

    use Extract;
    $obj = Extract->new->extract($FILENAME)->output;

=head1 DESCRIPTION

    ϥƥȤ顤̾׸Ф뤿Υ⥸塼롥
    ǥեȤǤϡϢˤٷ׻(׻ʿ)
    ԤʤѤƽ׸Ф롥

=head1 METHODS

    ʲΥ᥽åɤѲǽǤ롥Υ⥸塼˴ޤޤ
    ȾΥ᥽åɤϡϤ줿֥Ȥ֤褦߷פƤ롥
    ΤᡤʣΥ᥽åɤΥ󤬲ǽǤ롥

=head2 new

    󥹥ȥ饯᥽åɡ
     Extract ֥ȤΥ֥Ȥ֤

    usage : $obj = Extract->new;

=head2 extract

    ׸ФԤʤΥ᥽åǤϡʲΥ᥽åɤ
    ƤӽФԤʤäƤǤ롥

      1) calc_noun_imp : ̾ν٤׻
      2) window_func   : ؿŬѤˤ׸

    usage : $obj = $obj->extract($FILENAME); # ñΥե̾

=head2 output

    ̤Ϥ롥Ȥƥե̾Ϳ줿ˤϡ
    Υե˽񤭽ФԤʤʤϡɸϤ
    з̤ɽ롥

    ϤƤϡ١Ф줿̾졤и٤Ǥ롥

    usage : $obj = $obj->output;             # ɸϤ˽
            $obj = $obj->output("FILENAME"); # "FILENAME" ˽

=cut

;# ========================================================================
;# init -- to initialize variables
;#
;# usag: $obj->init([%PARAMETER]);
;#
;# ========================================================================
sub init ($;%) {
    my $this = shift;	# current object
    my %p    = @_;	# parameter
    my %env;


    $p{'juman'} =~ s/\s.+$// if (defined $p{'juman'}); # remove options
    $env{'sep'    } = $p{'sep'    } || 0;              # noun separator
    $env{'verbose'} = $p{'verbose'} || 0;              # verbose mode
    $env{'comb'   } = $p{'comb'   } || 0;              # noun combination
    $env{'rate'   } = $p{'rate'   } || 0.2;            # compound noun rate
    $env{'window' } = $p{'window' } || 5;              # window width
    $env{'min_frq'} = $p{'min_frq'} || 2;              # minimal frequency
    $env{'max_frq'} = $p{'max_frq'};                   # maximal frequency
    delete $this->{'noun_frq'} if $this->{'noun_frq'};
    $this->{'env'}  = \%env;
    return 1;
}

;# ------------------------------------------------------------------------
;# accessor methods
;# ------------------------------------------------------------------------


;# minimal noun frequency
sub min_frq ($;$) {
    defined $_[1]
	? $_[0]->{'env'}{'min_frq'} = $_[1]
	: return $_[0]->{'env'}{'min_frq'};
    delete $_[0]->{'ext_noun'};
    $_[0];
}

;# maximal noun frequency
sub max_frq ($;$) {
    defined $_[1]
	? $_[0]->{'env'}{'max_frq'} = $_[1]
	: return $_[0]->{'env'}{'max_frq'};
    delete $_[0]->{'ext_noun'};
    $_[0];
}

;# compound noun rate
sub rate ($;$) {
    defined $_[1]
	? $_[0]->{'env'}{'rate'} = $_[1]
	: return $_[0]->{'env'}{'rate'};
    delete $_[0]->{'ext_noun'};
    $_[0];
}

;# window width
sub window ($;$) {
    defined $_[1]
	? $_[0]->{'env'}{'window'} = $_[1]
	: return $_[0]->{'env'}{'window'};
    delete $_[0]->{'ext_noun'};
    $_[0];
}

;# single noun separator to show
sub sep ($;$) {
    defined $_[1]
	? $_[0]->{'env'}{'sep'} = $_[1]
	: return $_[0]->{'env'}{'sep'};
    $_[0];
}

;# verbose -- be verbose
;# 
;# $obj->verbose;    # verbose on
;# $obj->verbose(1); # verbose on
;# $obj->verbose(0); # verbose off
;#
sub verbose ($;$) {
    defined $_[1]
	? $_[0]->{'env'}{'verbose'} = $_[1]
	: $_[0]->{'env'}{'verbose'} = 1;
    $_[0];
}

;# comb -- think all single-noun combination or not
;# 
;# $obj->comb;       # comb on
;# $obj->comb(1);    # comb on
;# $obj->comb(0);    # comb off
;#
sub comb ($;$) {
    defined $_[1]
	? $_[0]->{'env'}{'comb'} = $_[1]
	: $_[0]->{'env'}{'comb'} = 1;
    $_[0];
}

;# extracted noun
sub extracted_noun ($) { $_[0]->{'ext_noun'}; }

;# ========================================================================
;# new -- constructor
;# 
;# usage: $obj = Extract->new();
;#
;# ========================================================================
sub new ($;%) {
    my $class = shift;
    my $this ={};
    bless $this, $class;
    $this->init(@_);
    $this;
}

;# ========================================================================
;# extract -- extract index words from input files given in args
;# 
;# usage: $obj = $obj->extract([\@FILES]);
;#
;# ========================================================================
sub extract ($;@) {
    my $this  = shift;		# object
    my @files = @_;		# input files (array ref)

    # show usage and die if no files are given
    $files[0] || $this->{'noun_frq'} || 
	die "usage: \$obj->extract(\@files)\;\n";

    # [1] calculate noun importance
    $this->calc_noun_imp()      unless $this->{'noun_imp'};

    # [2] extraction using window function
    $this->window_func()        unless $this->{'ext_noun'};

    $this;
}

;# ========================================================================
;# output -- output extracted noun
;# 
;# usage: $obj->output([$OUTPUT_FILE]);
;# 
;# ========================================================================
sub output ($;@) {
    my $this = shift;		# object
    my @files = @_;
    my $file = pop @files if @_[0] =~ /^\-/ && @files == 3 || @_[0] =~ /^[^\-]/ && @files == 2;		# output filename
    my $chr  = 1;		# dummy separate charactor
    open(OUT, ">>$file") || die "Cannot open $file. : $!\n" if $file;
    foreach my $n (@{$this->{'ext_noun'}}) {
	my $n_cp = $n;		# make a copy of the noun


	

	if ($this->{'env'}{'sep'} == 0) { # separate 1byte-char single noun
	    while ($n_cp =~ s/([\!-\~])\s([\!-\~])/$1\$chr$2/g) {};
##################################
	    $n_cp =~ s/\s//g;    #  ,,,  tuketa yametari
##################################
	    $n_cp =~ s/\$chr/ /g;

	} elsif ($this->{'env'}{'sep'} == 1) { # join 2byte-char single noun
	    while ($n_cp =~ s/([^\!-\~])\s([^\!-\~])/$1$2/g) {};

	} elsif ($this->{'env'}{'sep'} == 3) { # join all with "comma"
	    $n_cp =~ s/\s/\,/g;
	}

	printf
	   
 	     ("%-40s %20.2f %5d\n",
	      $n_cp, $this->{'noun_imp'}{$n}, $this->{'noun_frq'}{$n})
	    unless $file;

	printf OUT 

	    ("%-40s %20.2f %5d\n",
	     $n_cp, $this->{'noun_imp'}{$n}, $this->{'noun_frq'}{$n})
		if     $file;
    }
    close(OUT) if $file;
    $this =0;
}

;# ========================================================================
;# window_func -- Window function for extracted noun
;# 
;# usage: $obj = $obj->window_func();
;# 
;# ========================================================================
sub window_func ($) {
    my $this       = shift;		       # object
    my $n_frq      = $this->{'noun_frq'};      # noun frequency (hash ref)
    my $n_imp      = $this->{'noun_imp'};      # noun importance(hash ref)
    my $width      = $this->{'env'}{'window'}; # window width
    my $cn_rate    = $this->{'env'}{'rate'};   # compound noun rate
    my $half_width = int $width / 2;	       # half of window width
    my @extracted_nouns;	               # extracted nouns

    print STDERR "Window function ... " if $this->{'env'}{'verbose'};

    # sort nouns by importance
    my @noun_list = sort { $n_imp->{$b} <=> $n_imp->{$a} } keys %$n_imp;

    # window function
    foreach my $i (0 .. $#noun_list) {
	my $center_noun = $noun_list[$i];

	# --- "TOP" of noun list  ---
	# All nouns are extracted.
	if ($i < $half_width) {
	    # extract all noun
	    push(@extracted_nouns, $center_noun);
	    next;

	# --- "BOTTOM" of noun list ---
	# All nouns are not extraced.
	} elsif ($i > ($#noun_list + 1 - $half_width)) {
	    # void (do not extract)
	    next;

	# --- "MIDDLE" of noun list ---
	} else {
	    # count compound noun in the window
	    my $cmp_noun_count = 0;
	    foreach my $j (0 .. $width - 1) {
		$cmp_noun_count++
		    if ($noun_list[$i + $j - $half_width] =~ /\s/);
	    }
	    # extract the noun
	    push(@extracted_nouns, $center_noun)
		if ($cmp_noun_count > $width * $cn_rate);
	    next;
	}
	# end of window function
    }

    # extracted noun filter 
     @extracted_nouns =
 	grep { $n_frq->{$_} >= $this->{'env'}{'min_frq'} } @extracted_nouns;
     @extracted_nouns =
 	grep { $n_frq->{$_} <= $this->{'env'}{'max_frq'} } @extracted_nouns
 	    if $this->{'env'}{'max_frq'};

     print STDERR "done.\n"       if $this->{'env'}{'verbose'};
     print STDERR 
 	scalar @extracted_nouns,
 	" nouns are extracted.\n" if $this->{'env'}{'verbose'};

    $this->{'ext_noun'} = \@extracted_nouns;
    $this;
}

;# ========================================================================
;# calc_noun_imp - calculation for noun importance.
;# 
;# usage: $obj = $obj->calc_noun_imp();
;# 
;# ========================================================================
sub calc_noun_imp ($) {
    my $this = shift;		# object
    my %noun_imp;		# noun importance
    my %num_pre;		# number of continuation (pre  position)
    my %num_pst;		#                        (post position)
    my %work_flag;		# a flag for dull work

    print STDERR "Calculating importance  ... "
	if $this->{'env'}{'verbose'};

print STDERR
 	scalar keys %{$this->{'title'}} ; 

    # ---  Count continuation if $key is compound noun. ---
    foreach my $noun (sort { length $b <=> length $a }
		      keys %{$this->{'noun_frq'}}) {
	next if ($noun =~ /^\s*$/);
	count_continuation($noun, \%num_pre, \%num_pst, \%num_pre2, \%num_pst2,\%work_flag, $this->{'noun_frq'})
	    if ($noun =~ /\s/);
    }

    # --- Calculate importance of the noun. ---
    foreach my $noun (sort { length $b <=> length $a }
		      keys %{$this->{'noun_frq'}}) {
	next if ($noun =~ /^\s*$/);
	my $imp = imp_function($noun, \%num_pre, \%num_pst, \%num_pre2, \%num_pst2, $this->{'noun_frq'});

	# --- Set the importance value to %$noun_imp. ---
	$noun_imp{$noun} = $imp if ($imp);
    }
    print STDERR "done.\n" if $this->{'env'}{'verbose'};
    $this->{'noun_imp'} = \%noun_imp;
    $this;
}

;# ------------------------------------------------------------------------
;# imp_function -- A function for calculation of each noun's importance.
;# 
;# usage: $IMP = imp_function($NOUN, \%PRE, \%POST );
;#
sub imp_function ($\%\%\%\%\%) {
  
    my $cmp_noun = shift;	# a noun
    my $num_pre  = shift;	# number of single noun continuation (pre )
    my $num_pst  = shift;	#                                    (post)
    my $num_pre2  = shift;	# hash reference to count continuation (pre )
    my $num_pst2  = shift;	#                                      (post)
    my $frq      = shift;

    my $imp      = 1;		# noun importance

  
    $cmp_noun =~ s/\s\s/ /g;	        # remove 'no' from compound noun
    my @noun  = split(/\s/, $cmp_noun);	# @noun is a set of single nouns.

    # calculation
    foreach my $s_noun (@noun) {
	# $s_noun is a single noun.

	
	$imp *= ($num_pre2->{$s_noun} + 1) * ($num_pst2->{$s_noun} + 1);
	

	

    }

    $imp = $imp ** (1 / ( scalar @noun**1 )); 
    $imp = $imp * $frq->{$cmp_noun};

    return $imp;
}

;# ------------------------------------------------------------------------
;# count_continuation -- to count the number of single noun continuation
;# 
;# usage: count_continuation($COMPOUND_NOUN, \%PRE, \%POST, \%FLAG);
;# 
sub count_continuation ($\%\%\%\%\%\%) {
   
    my $cmp_noun = shift;	# a compound noun
    my $num_pre  = shift;	# hash reference to count continuation (pre )
    my $num_pst  = shift;	#                                      (post)
    my $num_pre2  = shift;	# hash reference to count continuation (pre )
    my $num_pst2  = shift;	#                                      (post)
    my $flag     = shift;	# hash reference of flag to count same
				# combination of single nouns only once
    my $frq      = shift;

    $cmp_noun =~ s/\s\s/ /g;	        # remove 'no' from compound noun
    my @noun  = split(/\s/, $cmp_noun);	# @noun is a set of single nouns.
 

    foreach my $i (0 .. $#noun) {
	my $pre_key = $noun[$i-1] . ' ' . $noun[$i]   if ($i > 0);
	my $pst_key = $noun[$i]   . ' ' . $noun[$i+1] if ($i < $#noun);


	if ($pre_key && $flag->{"pre $pre_key"} != 1) {
	    $num_pre->{$noun[$i]}++;
	    $flag->{"pre $pre_key"} = 1;
	}
	if ($pre_key ) {
	    $num_pre2->{$noun[$i]} += $frq->{$cmp_noun};
	}
	if ($pst_key && $flag->{"pst $pst_key"} != 1) {
	    $num_pst->{$noun[$i]}++; # 
	    $flag->{"pst $pst_key"} = 1;
	}			# 
	if ($pst_key ) {	# 
	    $num_pst2->{$noun[$i]} += $frq->{$cmp_noun};
	}
# initalize
	($pre_key, $pst_key) = ('', '');
    }

    #simpson D

    
    return 1;
}


;# ========================================================================
1;









